from .base import AuxService
import pandas as pd


def fact_mjd_to_datetime(fact_mjd):
    ''' convert fact mjds (days since unix epoch) to pandas datetimes '''
    return pd.to_datetime(fact_mjd * 24 * 3600 * 1e9)


__all__ = [
    'MagicWeather',
    'PfMini',
    'DriveTracking',
    'DrivePointing',
    'DriveSource',
    'FSCHumidity',
    'FSCTemperature',
    'FTMTriggerRates',
    'BiasVoltage',
    'FADTemperature',
]


class MagicWeather(AuxService):
    basename = 'MAGIC_WEATHER_DATA'
    renames = {
        'Time': 'timestamp',
        'T': 'temperature',
        'T_dew': 'dewpoint',
        'H': 'humidity',
        'P': 'pressure',
        'v': 'wind_speed',
        'v_max': 'wind_gust_speed',
        'd': 'wind_direction',
    }

    ignored_columns = ['stat', 'QoS']
    transforms = {'timestamp': fact_mjd_to_datetime}


class PfMini(AuxService):
    basename = 'PFMINI_CONTROL_DATA'
    renames = {
        'Time': 'timestamp',
        'Temperature': 'temperature',
        'Humidity': 'humidity',
    }

    ignored_columns = ['QoS', ]
    transforms = {'timestamp': fact_mjd_to_datetime}


class DriveTracking(AuxService):
    basename = 'DRIVE_CONTROL_TRACKING_POSITION'
    renames = {
        'Time': 'timestamp',
        'Ra': 'right_ascension',
        'Dec': 'declination',
        'Ha': 'hourangle',
        'SrcHa': 'hourangle_source',
        'SrcRa': 'right_ascension_source',
        'SrcDec': 'declination_source',
        'HaDec': 'hourangle_source',
        'Zd': 'zenith',
        'Az': 'azimuth',
        'dZd': 'zenith_deviation',
        'dAz': 'azimuth_deviation',
        'dev': 'absolute_control_deviation',
        'avgdev': 'average_control_deviation',
    }
    transforms = {'timestamp': fact_mjd_to_datetime}
    ignored_columns = ['QoS', ]


class DrivePointing(AuxService):
    basename = 'DRIVE_CONTROL_POINTING_POSITION'
    renames = {
        'Time': 'timestamp',
        'Zd': 'zenith',
        'Az': 'azimuth',
    }
    transforms = {'timestamp': fact_mjd_to_datetime}
    ignored_columns = ['QoS', ]


class DriveSource(AuxService):
    basename = 'DRIVE_CONTROL_SOURCE_POSITION'
    renames = {
        'Time': 'timestamp',
        'Ra_src': 'right_ascension_source',
        'Ra_cmd': 'right_ascension_command',
        'Dec_src': 'declination_source',
        'Dec_cmd': 'declination_command',
        'Offset': 'wobble_offset',
        'Angle': 'wobble_angle',
        'Name': 'source',
        'Period': 'orbit_period',
    }
    transforms = {'timestamp': fact_mjd_to_datetime}
    ignored_columns = ['QoS', ]


class FSCHumidity(AuxService):
    basename = 'FSC_CONTROL_HUMIDITY'
    renames = {
        'Time': 'timestamp',
        't': 'fsc_uptime',
        'H': 'humidity',
    }
    transforms = {'timestamp': fact_mjd_to_datetime}
    ignored_columns = ['QoS', ]


class FSCTemperature(AuxService):
    basename = 'FSC_CONTROL_TEMPERATURE'
    renames = {
        'Time': 'timestamp',
        't': 'fsc_uptime',
        'T_crate': 'crate_temperature',
        'T_sens': 'sensor_compartment_temperature',
        'T_ps': 'power_supply_temperature',
        'T_aux': 'auxiliary_power_supply_temperature',
        'T_back': 'ftm_backpanel_temperature',
        'T_eth': 'ethernet_temperature',
    }
    transforms = {'timestamp': fact_mjd_to_datetime}
    ignored_columns = ['QoS', ]


class FTMTriggerRates(AuxService):
    basename = 'FTM_CONTROL_TRIGGER_RATES'
    renames = {
        'Time': 'timestamp',
        'FTMtimeStamp': 'ftm_timestamp',
        'OnTimeCounter': 'effective_ontime',
        'TriggerCounter': 'trigger_counter',
        'TriggerRate': 'trigger_rate',
        'BoardRate': 'board_rate',
        'PatchRate': 'patch_rate',
        'OnTime': 'ontime',
        'ElapsedTime': 'elapsed_time',
    }
    transforms = {
        'timestamp': fact_mjd_to_datetime,
        'ftm_timestamp': lambda x: x/1e6,
        'effective_ontime': lambda x: x/1e6,
    }
    ignored_columns = ['QoS', ]


class BiasVoltage(AuxService):
    basename = "BIAS_CONTROL_VOLTAGE"
    renames = {
        'Time': 'timestamp',
        'Uout': 'bias_voltage',
    }
    transforms = {
        'timestamp': fact_mjd_to_datetime,
    }
    ignored_columns = ['QoS', ]


class FADTemperature(AuxService):
    basename = "FAD_CONTROL_TEMPERATURE"
    renames = {
        'Time': 'timestamp',
        'cnt': 'count',
        'temp': 'temperature',
    }
    transforms = {
        'timestamp': fact_mjd_to_datetime,
    }
    ignored_columns = ['QoS', ]



import matplotlib.pyplot as plt
from matplotlib.dates import DateFormatter
import matplotlib.transforms as mtransforms
from mpl_toolkits.axes_grid1.parasite_axes import SubplotHost
import numpy as np
from ..time import MJD_EPOCH


# Matplotlib uses ordinal for internal date represantion
# to get from ordinal to mjd, we shift by the ordinal value
# of the MJD_EPOCH
MJD_AXES_TRANSFORM = (
    mtransforms.Affine2D().translate(MJD_EPOCH.toordinal(), 0)
)


def create_datetime_mjd_axes(fig=None, *args, **kwargs):
    '''
    Create a plot with two x-axis, bottom axis using
    dates, top axis using mjd.

    Parameters
    ----------
    fig: matplotlib.Figure or None
        the figure to use, if None use plt.gcf()

    Returns
    -------
    ax: mpl_toolkits.axes_grid1.parasite_axes.SubplotHost
        The ax for the dates
    mjd_ax: mpl_toolkits.axes_grid1.parasite_axes.ParasiteAxis
        The axis with the mjd axis

    '''
    if fig is None:
        fig = plt.gcf()

    if args == []:
        ax = SubplotHost(fig, 1, 1, 1, **kwargs)
    else:
        ax = SubplotHost(fig, *args, **kwargs)

    # The second axis shows MJD if the first axis uses dates
    mjd_ax = ax.twin(MJD_AXES_TRANSFORM)
    mjd_ax.set_viewlim_mode('transform')

    # disable unwanted axes
    mjd_ax.axis['right'].toggle(ticklabels=False, ticks=False)
    mjd_ax.axis['bottom'].toggle(ticklabels=False, ticks=False)
    mjd_ax.axis['bottom'].toggle(label=False)

    # add/remove label
    mjd_ax.axis['top'].set_label('MJD')

    # Deactivate offset
    mjd_ax.ticklabel_format(useOffset=False)

    fig.add_subplot(ax)

    return ax, mjd_ax


def plot_excess_rate(binned_runs, outputfile=None, mjd=True):
    '''
    Create an excess rate plot from given data

    Parameters
    ----------
    binned_runs: pd.DataFrame
        Binned data as returned by `fact.analysis.bin_runs`
    outputfile: path
        if not None, the plot is saved to this file

    Returns
    -------
    ax_excess: matplotlib.axes.Axes
        the matplotlib Axes for the excess rate plot
    ax_significance: matplotlib.axes.Axes
        the matplotlib Axes for the significance plot
    '''
    fig = plt.figure()

    gridspec = plt.GridSpec(8, 1)
    spec_sig = gridspec.new_subplotspec((6, 0), rowspan=2)
    spec_rate = gridspec.new_subplotspec((1, 0), rowspan=5)

    ax_sig = plt.subplot(spec_sig)

    if mjd is True:
        ax_rate, ax_rate_mjd = create_datetime_mjd_axes(
            fig, spec_rate, sharex=ax_sig
        )
    else:
        ax_rate = plt.subplot(spec_rate, sharex=ax_sig)

    colors = [e['color'] for e in plt.rcParams['axes.prop_cycle']]

    labels = []
    plots = []

    groups = list(sorted(
        binned_runs.groupby('source'),
        key=lambda g: g[1].time_mean.min()
    ))

    for (name, group), color in zip(groups, colors):
        if len(group.index) == 0:
            continue

        labels.append(name)
        plots.append(ax_rate.errorbar(
            x=group.time_mean.values,
            y=group.excess_rate_per_h.values,
            xerr=group.time_width.values / 2,
            yerr=group.excess_rate_err.values,
            label=name,
            linestyle='',
            mec='none',
            color=color,
        ))

        ax_sig.errorbar(
            x=group.time_mean.values,
            y=group.significance.values,
            xerr=group.time_width.values / 2,
            label=name,
            linestyle='',
            mec='none',
            color=color,
        )

    fig.legend(
        plots,
        labels,
        loc='upper center',
        ncol=3,
        columnspacing=0.5,
        numpoints=1,
        handletextpad=0.1,
        bbox_to_anchor=[0.5, 0.99],
    )

    ax_rate.set_ylabel('Excess Event Rate / $\mathrm{h}^{-1}$')
    ax_sig.set_ylabel('$S_{\mathrm{Li/Ma}} \,\, / \,\, \sigma$')

    ymax = max(3.25, np.ceil(ax_sig.get_ylim()[1]))
    ax_sig.set_ylim(0, ymax)
    ax_sig.set_yticks(np.arange(0, ymax + 0.1, ymax // 4 + 1))

    plt.setp(ax_rate.get_xticklabels(), visible=False)
    plt.setp(ax_rate_mjd.get_xticklabels(), visible=False)
    plt.setp(ax_sig.get_xticklabels(), rotation=30, va='top', ha='right')

    ax_rate_mjd.set_xlabel('')

    if binned_runs.night.nunique() <= 7:
        ax_sig.xaxis.set_major_formatter(DateFormatter('%Y-%m-%d %H:%M'))
    else:
        ax_sig.xaxis.set_major_formatter(DateFormatter('%Y-%m-%d'))

    fig.tight_layout()

    if mjd is True:
        plt.setp(ax_rate_mjd.get_xticklabels(), visible=True)

    if outputfile is not None:
        fig.savefig(outputfile)

    if mjd is True:
        return ax_rate, ax_sig, ax_rate_mjd
    else:
        return ax_rate, ax_sig



from .camera_frame import CameraFrame
from .utils import equatorial_to_camera, camera_to_equatorial
from .utils import horizontal_to_camera, camera_to_horizontal


__all__ = [
    'CameraFrame',
    'equatorial_to_camera',
    'camera_to_equatorial',
    'horizontal_to_camera',
    'camera_to_horizontal',
]



import matplotlib.pyplot as plt

__all__ = ['calc_linewidth', 'calc_text_size']


def calc_linewidth(ax=None):
    """
    calculate the correct linewidth for the fact pixels,
    so that the patches fit nicely together

    Arguments
    ---------
    ax  : matplotlib Axes instance
        the axes you want to calculate the size for

    Returns
    -------
    linewidth : float
    """

    if ax is None:
        ax = plt.gca()

    fig = ax.get_figure()

    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width, height = bbox.width, bbox.height

    x1, x2 = ax.get_xlim()
    y1, y2 = ax.get_ylim()

    x_stretch = (x2 - x1) / 400
    y_stretch = (y2 - y1) / 400

    linewidth = min(width / x_stretch, height / y_stretch) / 10
    return linewidth


def calc_text_size(ax=None):
    if ax is None:
        ax = plt.gca()
    linewidth = calc_linewidth(ax)

    textsize = linewidth*5

    return textsize



import pandas as pd
import datetime
import numpy as np

from .statistics import li_ma_significance


def ontime_binning(runs, bin_width_minutes=20):
    '''
    Calculate bin numbers for given runs.
    A new bin is created if either a bin would have more ontime
    than `bin_width_minutes` or `run_start` of the next run is
    more than `bin_width_minutes` after `run_stop` of the last run.

    Parameters
    ----------

    runs: pd.DataFrame
        DataFrame containing analysis results and meta data
        for each run
    bin_width_minutes: number
        The desired amount of ontime in each bin.
        Note: The ontime in each bin will allways be
        slightly less than `bin_width_minutes`
    '''
    bin_width_sec = bin_width_minutes * 60
    bin_number = 0
    ontime_sum = 0

    bins = []
    last_stop = runs['run_start'].iloc[0]
    delta_t_max = datetime.timedelta(seconds=bin_width_sec)

    for key, row in runs.iterrows():
        delta_t = row.run_start - last_stop
        last_stop = row.run_stop

        if ontime_sum + row.ontime > bin_width_sec or delta_t > delta_t_max:
            bin_number += 1
            ontime_sum = 0

        bins.append(bin_number)
        ontime_sum += row.ontime

    return pd.Series(bins, index=runs.index)


def qla_binning(data, bin_width_minutes=20):
    '''
    The binning algorithm as used by lightcurve.c
    '''
    bin_number = 0
    ontime_sum = 0
    bins = []

    for key, row in data.iterrows():
        if ontime_sum + row.fOnTimeAfterCuts > bin_width_minutes * 60:
            bin_number += 1
            ontime_sum = 0

        bins.append(bin_number)
        ontime_sum += row['ontime']

    return pd.Series(bins, index=data.index)


def groupby_observation_blocks(runs):
    ''' Groupby for consecutive runs of the same source'''
    runs = runs.sort_values('run_start')
    new_source = runs.fSourceName != runs.fSourceName.shift(1)
    observation_blocks = new_source.cumsum()
    return runs.groupby(observation_blocks)


def nightly_binning(runs):
    nights = runs['night'].unique()
    bins = pd.Series(index=runs.index, dtype=int)

    for bin_id, night in enumerate(nights):
        bins.loc[runs.night == night] = bin_id

    return bins


def bin_runs(
        runs,
        alpha=0.2,
        binning_function=ontime_binning,
        **kwargs
        ):
    '''
    Bin runs using `binning_function` to assign bins to
    the individual runs.
    Calculates n_on, n_off, ontime, n_excess, excess_rate_per_h,
    excess_rate_err, li_ma_significance and bin_width

    Parameters
    ----------
    runs: pandas.DataFrame
        The analysis results and necessary metadata for each run.
        Required are: ontime, n_on, n_off, run_start, run_stop, source

    alpha: float
        The weight for the off regions, e.g. 1 / number of off regions

    binning_function: function
        A function that takes the run df and returns a
        pd.Series containing bin ids with the index of the origininal
        dataframe

    All `**kwargs` are passed to the binning function
    '''
    runs = runs.sort_values(by='run_start')
    sources = []
    for source, df in runs.groupby('source'):

        df = df.copy()
        df['bin'] = binning_function(df, **kwargs)

        binned = df.groupby('bin').aggregate({
            'ontime': 'sum',
            'n_on': 'sum',
            'n_off': 'sum',
            'run_start': 'min',
            'run_stop': 'max',
        })

        binned['n_excess'] = binned.n_on - binned.n_off * alpha
        binned['excess_rate_per_h'] = binned.n_excess / binned.ontime * 3600

        binned['time_width'] = binned.run_stop - binned.run_start
        binned['time_mean'] = binned.run_start + 0.5 * binned.time_width

        binned['excess_rate_err'] = np.sqrt(binned.n_on + alpha**2 * binned.n_off)
        binned['excess_rate_err'] /= binned.ontime / 3600

        binned['significance'] = li_ma_significance(
            binned.n_on, binned.n_off, 0.2
        )

        binned['source'] = source
        binned['night'] = (
            binned.time_mean - pd.Timedelta(hours=12)
        ).dt.strftime('%Y%m%d').astype(int)

        sources.append(binned)

    return pd.concat(sources)



from peewee import (
    Model,
    FloatField, IntegerField, DateTimeField, CharField, BigIntegerField,
    PrimaryKeyField, CompositeKey
)

from .database import factdata_db

class FactDataModel(Model):
    class Meta:
        database = factdata_db


class AnalysisResultsNightISDC(FactDataModel):
    fnight = IntegerField(db_column='fNight')
    fnumbgevts = FloatField(db_column='fNumBgEvts', null=True)
    fnumevtsafterbgcuts = IntegerField(db_column='fNumEvtsAfterBgCuts', null=True)
    fnumevtsaftercleaning = IntegerField(db_column='fNumEvtsAfterCleaning', null=True)
    fnumevtsafterqualcuts = IntegerField(db_column='fNumEvtsAfterQualCuts', null=True)
    fnumexcevts = FloatField(db_column='fNumExcEvts', null=True)
    fnumislandsmean = FloatField(db_column='fNumIslandsMean', null=True)
    fnumsigevts = FloatField(db_column='fNumSigEvts', null=True)
    fontimeaftercuts = FloatField(db_column='fOnTimeAfterCuts', null=True)
    fsourcekey = IntegerField(db_column='fSourceKey')

    class Meta:
        db_table = 'AnalysisResultsNightISDC'
        indexes = (
            (('fsourcekey', 'fnight'), True),
        )
        primary_key = CompositeKey('fnight', 'fsourcekey')


class AnalysisResultsNightLP(FactDataModel):
    fnight = IntegerField(db_column='fNight')
    fnumbgevts = FloatField(db_column='fNumBgEvts', null=True)
    fnumevtsafterbgcuts = IntegerField(db_column='fNumEvtsAfterBgCuts', null=True)
    fnumevtsaftercleaning = IntegerField(db_column='fNumEvtsAfterCleaning', null=True)
    fnumevtsafterqualcuts = IntegerField(db_column='fNumEvtsAfterQualCuts', null=True)
    fnumexcevts = FloatField(db_column='fNumExcEvts', null=True)
    fnumislandsmean = FloatField(db_column='fNumIslandsMean', null=True)
    fnumsigevts = FloatField(db_column='fNumSigEvts', null=True)
    fontimeaftercuts = FloatField(db_column='fOnTimeAfterCuts', null=True)
    fsourcekey = IntegerField(db_column='fSourceKey')

    class Meta:
        db_table = 'AnalysisResultsNightLP'
        indexes = (
            (('fsourcekey', 'fnight'), True),
        )
        primary_key = CompositeKey('fnight', 'fsourcekey')


class AnalysisResultsRunISDC(FactDataModel):
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fnumbgevts = FloatField(db_column='fNumBgEvts', null=True)
    fnumevtsafterantithetacut = FloatField(db_column='fNumEvtsAfterAntiThetaCut')
    fnumevtsafterbgcuts = IntegerField(db_column='fNumEvtsAfterBgCuts', null=True)
    fnumevtsaftercleaning = IntegerField(db_column='fNumEvtsAfterCleaning', null=True)
    fnumevtsafterqualcuts = IntegerField(db_column='fNumEvtsAfterQualCuts', null=True)
    fnumexcevts = FloatField(db_column='fNumExcEvts', null=True)
    fnumislandsmean = FloatField(db_column='fNumIslandsMean', null=True)
    fnumsigevts = FloatField(db_column='fNumSigEvts', null=True)
    fontimeaftercuts = FloatField(db_column='fOnTimeAfterCuts', null=True)
    frunid = IntegerField(db_column='fRunID')

    class Meta:
        db_table = 'AnalysisResultsRunISDC'
        indexes = (
            (('frunid', 'fnight'), True),
        )
        primary_key = CompositeKey('fnight', 'frunid')


class AnalysisResultsRunLP(FactDataModel):
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fnumbgevts = FloatField(db_column='fNumBgEvts', null=True)
    fnumevtsafterbgcuts = IntegerField(db_column='fNumEvtsAfterBgCuts', null=True)
    fnumevtsaftercleaning = IntegerField(db_column='fNumEvtsAfterCleaning', null=True)
    fnumevtsafterqualcuts = IntegerField(db_column='fNumEvtsAfterQualCuts', null=True)
    fnumexcevts = FloatField(db_column='fNumExcEvts', null=True)
    fnumislandsmean = FloatField(db_column='fNumIslandsMean', null=True)
    fnumsigevts = FloatField(db_column='fNumSigEvts', null=True)
    fontimeaftercuts = FloatField(db_column='fOnTimeAfterCuts', null=True)
    frunid = IntegerField(db_column='fRunID')

    class Meta:
        db_table = 'AnalysisResultsRunLP'
        indexes = (
            (('frunid', 'fnight'), True),
        )
        primary_key = CompositeKey('fnight', 'frunid')


class AutoSchedule(FactDataModel):
    fdata = CharField(db_column='fData', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fmeasurementid = IntegerField(db_column='fMeasurementID')
    fmeasurementtypekey = IntegerField(db_column='fMeasurementTypeKey')
    fscheduleid = PrimaryKeyField(db_column='fScheduleID')
    fsourcekey = IntegerField(db_column='fSourceKey', null=True)
    fstart = DateTimeField(db_column='fStart')
    fuser = CharField(db_column='fUser')

    class Meta:
        db_table = 'AutoSchedule'
        indexes = (
            (('fstart', 'fmeasurementid'), True),
        )


class AuxDataInsertStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = PrimaryKeyField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'AuxDataInsertStatus'


class AuxFilesAvailISDCStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = PrimaryKeyField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'AuxFilesAvailISDCStatus'


class Calibration(FactDataModel):
    fcaldev = FloatField(db_column='fCalDev', null=True)
    fcalerrdev = FloatField(db_column='fCalErrDev', null=True)
    fcalerrmed = FloatField(db_column='fCalErrMed', null=True)
    fcalmed = FloatField(db_column='fCalMed', null=True)
    fcalvstmdev = FloatField(db_column='fCalVsTmDev', null=True)
    fcalvstmmean = FloatField(db_column='fCalVsTmMean', null=True)
    fcalvstmmed = FloatField(db_column='fCalVsTmMed', null=True)
    fcalvstmrms = FloatField(db_column='fCalVsTmRms', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fpedextdev = FloatField(db_column='fPedExtDev', null=True)
    fpedexterrdev = FloatField(db_column='fPedExtErrDev', null=True)
    fpedexterrmed = FloatField(db_column='fPedExtErrMed', null=True)
    fpedextmed = FloatField(db_column='fPedExtMed', null=True)
    fpedrndmdev = FloatField(db_column='fPedRndmDev', null=True)
    fpedrndmerrdev = FloatField(db_column='fPedRndmErrDev', null=True)
    fpedrndmerrmed = FloatField(db_column='fPedRndmErrMed', null=True)
    fpedrndmmed = FloatField(db_column='fPedRndmMed', null=True)
    frunstart = DateTimeField(db_column='fRunStart', null=True)
    fsequenceid = IntegerField(db_column='fSequenceID')
    ftmcaldev = FloatField(db_column='fTmCalDev', null=True)
    ftmcalerrdev = FloatField(db_column='fTmCalErrDev', null=True)
    ftmcalerrmed = FloatField(db_column='fTmCalErrMed', null=True)
    ftmcalmed = FloatField(db_column='fTmCalMed', null=True)
    ftmvstmdev = FloatField(db_column='fTmVsTmDev', null=True)
    ftmvstmmean = FloatField(db_column='fTmVsTmMean', null=True)
    ftmvstmmed = FloatField(db_column='fTmVsTmMed', null=True)
    ftmvstmrms = FloatField(db_column='fTmVsTmRms', null=True)
    fzenithdistance = IntegerField(db_column='fZenithDistance', null=True)

    class Meta:
        db_table = 'Calibration'
        indexes = (
            (('fnight', 'fsequenceid'), True),
        )
        primary_key = CompositeKey('fnight', 'fsequenceid')


class CalibrationInfo(FactDataModel):
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fsequenceid = IntegerField(db_column='fSequenceID')

    class Meta:
        db_table = 'CalibrationInfo'
        indexes = (
            (('fnight', 'fsequenceid'), True),
        )
        primary_key = CompositeKey('fnight', 'fsequenceid')


class CallistoStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fsequenceid = IntegerField(db_column='fSequenceID')
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'CallistoStatus'
        indexes = (
            (('fnight', 'fsequenceid'), True),
        )
        primary_key = CompositeKey('fnight', 'fsequenceid')


class CallistoWueStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fsequenceid = IntegerField(db_column='fSequenceID')
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'CallistoWueStatus'
        indexes = (
            (('fnight', 'fsequenceid'), True),
        )
        primary_key = CompositeKey('fnight', 'fsequenceid')


class DriveFileAvailISDCStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = PrimaryKeyField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'DriveFileAvailISDCStatus'


class DriveFileAvailWueStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = PrimaryKeyField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'DriveFileAvailWueStatus'


class ExcludedFDA(FactDataModel):
    fexcludedfda = CharField(db_column='fExcludedFDA', unique=True)
    fexcludedfdakey = PrimaryKeyField(db_column='fExcludedFDAKEY')
    fexcludedfdaname = CharField(db_column='fExcludedFDAName', unique=True)

    class Meta:
        db_table = 'ExcludedFDA'


class FillCalibStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fsequenceid = IntegerField(db_column='fSequenceID')
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'FillCalibStatus'
        indexes = (
            (('fnight', 'fsequenceid'), True),
        )
        primary_key = CompositeKey('fnight', 'fsequenceid')


class FillStarStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fsequenceid = IntegerField(db_column='fSequenceID')
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'FillStarStatus'
        indexes = (
            (('fnight', 'fsequenceid'), True),
        )
        primary_key = CompositeKey('fnight', 'fsequenceid')


class MeasurementType(FactDataModel):
    fisunlimited = IntegerField(db_column='fIsUnlimited')
    fmeasurementtype = CharField(db_column='fMeasurementType')
    fmeasurementtypekey = PrimaryKeyField(db_column='fMeasurementTypeKey')
    fmeasurementtypename = CharField(db_column='fMeasurementTypeName', unique=True)
    fneedssource = IntegerField(db_column='fNeedsSource')

    class Meta:
        db_table = 'MeasurementType'


class ObservationTimes(FactDataModel):
    fmjd = IntegerField(db_column='fMjd')
    fmoonrise = DateTimeField(db_column='fMoonRise')
    fmoonset = DateTimeField(db_column='fMoonSet')
    fnight = PrimaryKeyField(db_column='fNight')
    fnumdarkhours = FloatField(db_column='fNumDarkHours')
    fstartdarknight = DateTimeField(db_column='fStartDarkNight')
    fstartdarktime = DateTimeField(db_column='fStartDarkTime')
    fstartobservation = DateTimeField(db_column='fStartObservation')
    fstopdarknight = DateTimeField(db_column='fStopDarkNight')
    fstopdarktime = DateTimeField(db_column='fStopDarkTime')
    fstopobservation = DateTimeField(db_column='fStopObservation')

    class Meta:
        db_table = 'ObservationTimes'


class Observatory(FactDataModel):
    fmagnetbx = FloatField(db_column='fMagnetBX', null=True)
    fmagnetbz = FloatField(db_column='fMagnetBZ', null=True)
    fmagnetrotation = FloatField(db_column='fMagnetRotation', null=True)
    fobslevel = IntegerField(db_column='fObsLevel')
    fobservatory = CharField(db_column='fObservatory')
    fobservatorykey = PrimaryKeyField(db_column='fObservatoryKEY')
    fobservatoryname = CharField(db_column='fObservatoryName')

    class Meta:
        db_table = 'Observatory'


class RatesFileAvailISDCStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = PrimaryKeyField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'RatesFileAvailISDCStatus'


class RatesFileAvailWueStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = PrimaryKeyField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'RatesFileAvailWueStatus'


class RateScan(FactDataModel):
    fazmax = FloatField(db_column='fAzMax')
    fazmin = FloatField(db_column='fAzMin')
    fcurrentmedmean = FloatField(db_column='fCurrentMedMean', null=True)
    fdecmean = FloatField(db_column='fDecMean')
    fnight = IntegerField(db_column='fNight')
    fnumpoints = IntegerField(db_column='fNumPoints', null=True)
    fovervoltage = FloatField(db_column='fOvervoltage', null=True)
    framean = FloatField(db_column='fRaMean')
    fratebegin = FloatField(db_column='fRateBegin')
    frateend = FloatField(db_column='fRateEnd')
    fratescanid = PrimaryKeyField(db_column='fRatescanID')
    fthresholdbegin = IntegerField(db_column='fThresholdBegin')
    fthresholdend = IntegerField(db_column='fThresholdEnd')
    ftimebegin = DateTimeField(db_column='fTimeBegin')
    ftimeend = DateTimeField(db_column='fTimeEnd')
    fvoltageison = IntegerField(db_column='fVoltageIsOn', null=True)
    fzdmax = FloatField(db_column='fZdMax')
    fzdmin = FloatField(db_column='fZdMin')

    class Meta:
        db_table = 'Ratescan'


class RawFileAvailISDCStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    frunid = IntegerField(db_column='fRunID')
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'RawFileAvailISDCStatus'
        indexes = (
            (('fnight', 'frunid'), True),
        )
        primary_key = CompositeKey('fnight', 'frunid')


class RawFileAvailWueStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    frunid = IntegerField(db_column='fRunID')
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'RawFileAvailWueStatus'
        indexes = (
            (('fnight', 'frunid'), True),
        )
        primary_key = CompositeKey('fnight', 'frunid')


class RawfileRsyncedISDCStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    frunid = IntegerField(db_column='fRunID')
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'RawFileRsyncedISDCStatus'
        indexes = (
            (('fnight', 'frunid'), True),
        )
        primary_key = CompositeKey('fnight', 'frunid')


class RunComments(FactDataModel):
    fcomment = CharField(db_column='fComment', null=True)
    fcommentkey = PrimaryKeyField(db_column='fCommentKEY')
    fnight = IntegerField(db_column='fNight')
    frunid = IntegerField(db_column='fRunID')
    fuser = CharField(db_column='fUser', null=True)

    class Meta:
        db_table = 'RunComments'


class RunInfo(FactDataModel):
    fangletomoon = FloatField(db_column='fAngleToMoon', null=True)
    fangletosun = FloatField(db_column='fAngleToSun', null=True)
    fazimuthmax = FloatField(db_column='fAzimuthMax', null=True)
    fazimuthmean = FloatField(db_column='fAzimuthMean', null=True)
    fazimuthmin = FloatField(db_column='fAzimuthMin', null=True)
    fbiasvoltagemedian = FloatField(db_column='fBiasVoltageMedian', null=True)
    fcamhumiditymean = FloatField(db_column='fCamHumidityMean', null=True)
    fcameratempmean = FloatField(db_column='fCameraTempMean', null=True)
    fcameratemprms = FloatField(db_column='fCameraTempRms', null=True)
    fcameratemprmsmean = FloatField(db_column='fCameraTempRmsMean', null=True)
    fchecksum = CharField(db_column='fCheckSum', null=True)
    fcompiletime = DateTimeField(db_column='fCompileTime', null=True)
    fcontainertempmean = FloatField(db_column='fContainerTempMean', null=True)
    fctrldevmean = FloatField(db_column='fCtrlDevMean', null=True)
    fctrldevrms = FloatField(db_column='fCtrlDevRms', null=True)
    fcurrentsdevmean = FloatField(db_column='fCurrentsDevMean', null=True)
    fcurrentsdevrms = FloatField(db_column='fCurrentsDevRms', null=True)
    fcurrentsdifftoprediction = FloatField(db_column='fCurrentsDiffToPrediction', null=True)
    fcurrentslinerms = FloatField(db_column='fCurrentsLineRms', null=True)
    fcurrentsmedmean = FloatField(db_column='fCurrentsMedMean', null=True)
    fcurrentsmedmeanbeg = FloatField(db_column='fCurrentsMedMeanBeg', null=True)
    fcurrentsmedmeanend = FloatField(db_column='fCurrentsMedMeanEnd', null=True)
    fcurrentsmedrms = FloatField(db_column='fCurrentsMedRms', null=True)
    fcurrentsreldifftoprediction = FloatField(db_column='fCurrentsRelDiffToPrediction', null=True)
    fcurrentsrellinerms = FloatField(db_column='fCurrentsRelLineRms', null=True)
    fdatasum = CharField(db_column='fDataSum', null=True)
    fdeclination = FloatField(db_column='fDeclination', null=True)
    fdrsstep = IntegerField(db_column='fDrsStep', null=True)
    fdrstempmaxmean = FloatField(db_column='fDrsTempMaxMean', null=True)
    fdrstempmaxrmsmean = FloatField(db_column='fDrsTempMaxRmsMean', null=True)
    fdrstempminmean = FloatField(db_column='fDrsTempMinMean', null=True)
    fdrstempminrmsmean = FloatField(db_column='fDrsTempMinRmsMean', null=True)
    feffectiveon = FloatField(db_column='fEffectiveOn', null=True)
    feffectiveonrms = FloatField(db_column='fEffectiveOnRms', null=True)
    fexcludedfdakey = IntegerField(db_column='fExcludedFDAKEY', null=True)
    ffilesize = BigIntegerField(db_column='fFileSize', null=True)
    ffitsfileerrors = IntegerField(db_column='fFitsFileErrors')
    fhasdrsfile = IntegerField(db_column='fHasDrsFile')
    flastupdate = DateTimeField(db_column='fLastUpdate')
    flidartransmission12 = FloatField(db_column='fLidarTransmission12', null=True)
    flidartransmission3 = FloatField(db_column='fLidarTransmission3', null=True)
    flidartransmission6 = FloatField(db_column='fLidarTransmission6', null=True)
    flidartransmission9 = FloatField(db_column='fLidarTransmission9', null=True)
    fmd5sumraw = CharField(db_column='fMd5sumRaw', null=True)
    fmd5sumrawzip = CharField(db_column='fMd5sumRawZip', null=True)
    fmoondisk = FloatField(db_column='fMoonDisk', null=True)
    fmoonzenithdistance = FloatField(db_column='fMoonZenithDistance', null=True)
    fnight = IntegerField(db_column='fNight')
    fnumelptrigger = IntegerField(db_column='fNumELPTrigger', null=True)
    fnumevents = IntegerField(db_column='fNumEvents', null=True)
    fnumext1trigger = IntegerField(db_column='fNumExt1Trigger', null=True)
    fnumext2trigger = IntegerField(db_column='fNumExt2Trigger', null=True)
    fnumilptrigger = IntegerField(db_column='fNumILPTrigger', null=True)
    fnumothertrigger = IntegerField(db_column='fNumOtherTrigger', null=True)
    fnumpedestaltrigger = IntegerField(db_column='fNumPedestalTrigger', null=True)
    fnumphysicstrigger = IntegerField(db_column='fNumPhysicsTrigger', null=True)
    fnumtimetrigger = IntegerField(db_column='fNumTimeTrigger', null=True)
    fontime = FloatField(db_column='fOnTime', null=True)
    foutsidetempmean = FloatField(db_column='fOutsideTempMean', null=True)
    foutsidetemprms = FloatField(db_column='fOutsideTempRms', null=True)
    fperiod = IntegerField(db_column='fPeriod', null=True)
    froi = IntegerField(db_column='fROI')
    froitimemarker = IntegerField(db_column='fROITimeMarker', null=True)
    frevisionnumber = CharField(db_column='fRevisionNumber', null=True)
    frightascension = FloatField(db_column='fRightAscension', null=True)
    frunid = IntegerField(db_column='fRunID')
    frunstart = DateTimeField(db_column='fRunStart', null=True)
    frunstop = DateTimeField(db_column='fRunStop', null=True)
    fruntypekey = IntegerField(db_column='fRunTypeKey')
    fsequenceid = IntegerField(db_column='fSequenceID', null=True)
    fsourcekey = IntegerField(db_column='fSourceKEY', null=True)
    fsqmmaglinfitchi2 = FloatField(db_column='fSqmMagLinFitChi2', null=True)
    fsqmmaglinfitndf = IntegerField(db_column='fSqmMagLinFitNdf', null=True)
    fsqmmaglinfitpvalue = FloatField(db_column='fSqmMagLinFitPValue', null=True)
    fsqmmaglinfitslope = FloatField(db_column='fSqmMagLinFitSlope', null=True)
    fsqmmagmean = FloatField(db_column='fSqmMagMean', null=True)
    fsunzenithdistance = FloatField(db_column='fSunZenithDistance', null=True)
    ftngdust = FloatField(db_column='fTNGDust', null=True)
    fthresholdavgmean = FloatField(db_column='fThresholdAvgMean', null=True)
    fthresholdmax = IntegerField(db_column='fThresholdMax', null=True)
    fthresholdmedmean = FloatField(db_column='fThresholdMedMean', null=True)
    fthresholdmedrms = FloatField(db_column='fThresholdMedRms', null=True)
    fthresholdmedian = FloatField(db_column='fThresholdMedian', null=True)
    fthresholdminset = IntegerField(db_column='fThresholdMinSet', null=True)
    fthresholdmintimediff = IntegerField(db_column='fThresholdMinTimeDiff', null=True)
    ftriggerratemedian = FloatField(db_column='fTriggerRateMedian', null=True)
    ftriggerraterms = FloatField(db_column='fTriggerRateRms', null=True)
    ftriggerratetimeover100 = FloatField(db_column='fTriggerRateTimeOver100', null=True)
    ftriggerratetimeover125 = FloatField(db_column='fTriggerRateTimeOver125', null=True)
    ftriggerratetimeover150 = FloatField(db_column='fTriggerRateTimeOver150', null=True)
    ftriggerratetimeover175 = FloatField(db_column='fTriggerRateTimeOver175', null=True)
    fzenithdistancemax = FloatField(db_column='fZenithDistanceMax', null=True)
    fzenithdistancemean = FloatField(db_column='fZenithDistanceMean', null=True)
    fzenithdistancemin = FloatField(db_column='fZenithDistanceMin', null=True)

    class Meta:
        db_table = 'RunInfo'
        indexes = (
            (('fnight', 'frunid'), True),
        )
        primary_key = CompositeKey('fnight', 'frunid')


class RunType(FactDataModel):
    fruntype = CharField(db_column='fRunType', unique=True)
    fruntypekey = PrimaryKeyField(db_column='fRunTypeKEY')
    fruntypename = CharField(db_column='fRunTypeName', unique=True)

    class Meta:
        db_table = 'RunType'


class Schedule(FactDataModel):
    fdata = CharField(db_column='fData', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fmeasurementid = IntegerField(db_column='fMeasurementID')
    fmeasurementtypekey = IntegerField(db_column='fMeasurementTypeKey')
    fscheduleid = PrimaryKeyField(db_column='fScheduleID')
    fsourcekey = IntegerField(db_column='fSourceKey', null=True)
    fstart = DateTimeField(db_column='fStart')
    fuser = CharField(db_column='fUser')

    class Meta:
        db_table = 'Schedule'
        indexes = (
            (('fstart', 'fmeasurementid'), True),
        )


class SequenceComments(FactDataModel):
    fcomment = CharField(db_column='fComment', null=True)
    fcommentkey = PrimaryKeyField(db_column='fCommentKEY')
    fnight = IntegerField(db_column='fNight')
    fsequenceid = IntegerField(db_column='fSequenceID')
    fuser = CharField(db_column='fUser', null=True)

    class Meta:
        db_table = 'SequenceComments'


class SequenceFileAvailISDCStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fsequenceid = IntegerField(db_column='fSequenceID')
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'SequenceFileAvailISDCStatus'
        indexes = (
            (('fnight', 'fsequenceid'), True),
        )
        primary_key = CompositeKey('fnight', 'fsequenceid')


class SequenceFileAVAILWueStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fsequenceid = IntegerField(db_column='fSequenceID')
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'SequenceFileAvailWueStatus'
        indexes = (
            (('fnight', 'fsequenceid'), True),
        )
        primary_key = CompositeKey('fnight', 'fsequenceid')


class SequenceInfo(FactDataModel):
    fazimuthmax = FloatField(db_column='fAzimuthMax', null=True)
    fazimuthmin = FloatField(db_column='fAzimuthMin', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fnumevents = IntegerField(db_column='fNumEvents')
    fsequenceduration = IntegerField(db_column='fSequenceDuration')
    fsequenceid = IntegerField(db_column='fSequenceID')
    fsequencestart = DateTimeField(db_column='fSequenceStart')
    fsequencestop = DateTimeField(db_column='fSequenceStop')
    fsourcekey = IntegerField(db_column='fSourceKEY')
    fzenithdistancemax = FloatField(db_column='fZenithDistanceMax', null=True)
    fzenithdistancemin = FloatField(db_column='fZenithDistanceMin', null=True)

    class Meta:
        db_table = 'SequenceInfo'
        indexes = (
            (('fnight', 'fsequenceid'), True),
        )
        primary_key = CompositeKey('fnight', 'fsequenceid')


class Source(FactDataModel):
    fdeclination = FloatField(db_column='fDeclination')
    fepochkey = IntegerField(db_column='fEpochKEY', null=True)
    fflux = FloatField(db_column='fFlux', null=True)
    fistoo = IntegerField(db_column='fIsToO')
    fmagnitude = FloatField(db_column='fMagnitude', null=True)
    frightascension = FloatField(db_column='fRightAscension')
    fslope = FloatField(db_column='fSlope', null=True)
    fsourcekey = PrimaryKeyField(db_column='fSourceKEY')
    fsourcename = CharField(db_column='fSourceName')
    fsourcetypekey = IntegerField(db_column='fSourceTypeKey', index=True)
    fwobbleangle0 = IntegerField(db_column='fWobbleAngle0')
    fwobbleangle1 = IntegerField(db_column='fWobbleAngle1')
    fwobbleoffset = FloatField(db_column='fWobbleOffset')

    class Meta:
        db_table = 'Source'


class SourceType(FactDataModel):
    fsourcetype = CharField(db_column='fSourceType', unique=True)
    fsourcetypekey = PrimaryKeyField(db_column='fSourceTypeKEY')
    fsourcetypename = CharField(db_column='fSourceTypeName', unique=True)

    class Meta:
        db_table = 'SourceType'


class StarInfo(FactDataModel):
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fsequenceid = IntegerField(db_column='fSequenceID')

    class Meta:
        db_table = 'StarInfo'
        indexes = (
            (('fnight', 'fsequenceid'), True),
        )
        primary_key = CompositeKey('fnight', 'fsequenceid')


class StarStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fsequenceid = IntegerField(db_column='fSequenceID')
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'StarStatus'
        indexes = (
            (('fnight', 'fsequenceid'), True),
        )
        primary_key = CompositeKey('fnight', 'fsequenceid')


class StarWueStatus(FactDataModel):
    favailable = DateTimeField(db_column='fAvailable', null=True)
    flastupdate = DateTimeField(db_column='fLastUpdate')
    fnight = IntegerField(db_column='fNight')
    fpriority = IntegerField(db_column='fPriority')
    fprocessingsitekey = IntegerField(db_column='fProcessingSiteKEY', null=True)
    freturncode = IntegerField(db_column='fReturnCode', null=True)
    fsequenceid = IntegerField(db_column='fSequenceID')
    fstarttime = DateTimeField(db_column='fStartTime', null=True)
    fstoptime = DateTimeField(db_column='fStopTime', null=True)

    class Meta:
        db_table = 'StarWueStatus'
        indexes = (
            (('fnight', 'fsequenceid'), True),
        )
        primary_key = CompositeKey('fnight', 'fsequenceid')


class TPointData(FactDataModel):
    fazencoder = FloatField(db_column='fAzEncoder')
    fazimage = FloatField(db_column='fAzImage')
    faznominal = FloatField(db_column='fAzNominal')
    fdeclination = FloatField(db_column='fDeclination')
    fintensity = FloatField(db_column='fIntensity')
    fmagcatalog = FloatField(db_column='fMagCatalog')
    fmagmeasured = FloatField(db_column='fMagMeasured')
    fnumleds = IntegerField(db_column='fNumLEDs')
    fnumrings = IntegerField(db_column='fNumRings')
    fqualitykey = IntegerField(db_column='fQualityKEY')
    frightascension = FloatField(db_column='fRightAscension')
    fstarname = CharField(db_column='fStarName')
    ftpointkey = PrimaryKeyField(db_column='fTPointKey')
    ftimestamp = DateTimeField(db_column='fTimeStamp')
    fxposcenter = FloatField(db_column='fXPosCenter')
    fxposstar = FloatField(db_column='fXPosStar')
    fyposcenter = FloatField(db_column='fYPosCenter')
    fyposstar = FloatField(db_column='fYPosStar')
    fzdencoder = FloatField(db_column='fZdEncoder')
    fzdimage = FloatField(db_column='fZdImage')
    fzdnominal = FloatField(db_column='fZdNominal')

    class Meta:
        db_table = 'TPointData'


class TPointQuality(FactDataModel):
    fquality = CharField(db_column='fQuality', null=True, unique=True)
    fqualitykey = PrimaryKeyField(db_column='fQualityKEY')
    fqualityname = CharField(db_column='fQualityName', null=True, unique=True)

    class Meta:
        db_table = 'TPointQuality'



# coding: utf-8
import pandas as pd

from joblib import Parallel, delayed
import h5py
import click

from ...coordinates.utils import camera_to_equatorial
from ...io import read_h5py_chunked
from ...io import create_empty_h5py_dataset, append_to_h5py_dataset


def calc_ra_dec(events):
    events['obstime'] = pd.to_datetime(events['timestamp'])

    events['ra_prediction'], events['dec_prediction'] = camera_to_equatorial(
        events['source_x_prediction'],
        events['source_y_prediction'],
        events['pointing_position_zd'],
        events['pointing_position_az'],
        events['obstime'].dt.to_pydatetime(),
    )
    return events


columns = ('ra_prediction', 'dec_prediction')


@click.command()
@click.argument('INPUTFILE', type=click.Path(exists=True, dir_okay=False))
@click.option('-c', '--chunksize', type=int, default=10000)
@click.option('-n', '--n-jobs', type=int, default=-1)
@click.option('-y', '--yes', is_flag=True, help='Do not ask to overwrite existing keys')
def main(inputfile, chunksize, n_jobs, yes):
    '''
    Calculate ra and dec from source position in camera coordinates
    e.g. for example for files analysed with the classifier-tools

    The following keys have to be present in the h5py hdf5 file.
        * pointing_position_az
        * pointing_position_zd
        * source_x_prediction
        * source_y_prediction
        * unix_time_utc
    '''

    with h5py.File(inputfile, 'r') as f:
        if any(col in f['events'].keys() for col in columns) and not yes:
            click.confirm('Output keys already exist, overwrite? ', abort=True)

    df_it = read_h5py_chunked(
        inputfile,
        key='events',
        columns=[
            'pointing_position_az',
            'pointing_position_zd',
            'source_x_prediction',
            'source_y_prediction',
            'timestamp',
        ],
        chunksize=chunksize
    )
    with Parallel(n_jobs, verbose=10) as pool:

        dfs = pool(
            delayed(calc_ra_dec)(df)
            for df, start, stop in df_it
        )

    df = pd.concat(dfs)

    with h5py.File(inputfile, mode='r+') as f:
        for col in ('ra_prediction', 'dec_prediction'):
            if col in f['events']:
                del f['events'][col]
            create_empty_h5py_dataset(df[col].values, f['events'], col)
            append_to_h5py_dataset(df[col].values, f['events'][col])


if __name__ == '__main__':
    main()



from astropy.coordinates import AltAz, ICRS, SkyCoord
from astropy.time import Time
import astropy.units as u
import numpy as np
import pandas as pd
from .camera_frame import CameraFrame
from ..instrument.constants import LOCATION
from datetime import datetime


def arrays_to_altaz(zenith, azimuth, obstime=None):
    if obstime is not None:
        obstime = to_astropy_time(obstime)
    return AltAz(
        az=np.asanyarray(azimuth) * u.deg,
        alt=np.asanyarray(90 - zenith) * u.deg,
        location=LOCATION,
        obstime=obstime,
    )


def arrays_to_camera(x, y, pointing_direction, obstime=None, rotated=True):
    if obstime is not None:
        obstime = to_astropy_time(obstime)
    frame = CameraFrame(
        pointing_direction=pointing_direction,
        obstime=obstime,
        rotated=rotated,
    )
    return SkyCoord(
        x=np.asanyarray(x) * u.mm,
        y=np.asanyarray(y) * u.mm,
        frame=frame,
    )


def arrays_to_equatorial(ra, dec, obstime=None):
    if obstime is not None:
        obstime = to_astropy_time(obstime)

    return SkyCoord(
        ra=np.asanyarray(ra) * u.hourangle,
        dec=np.asanyarray(dec) * u.deg,
        obstime=obstime
    )


def to_astropy_time(series_or_array):
    '''
    Convert a pandas or numpy time object to an astropy time
    '''
    if isinstance(series_or_array, Time):
        return series_or_array
    elif isinstance(series_or_array, pd.Series):
        time = series_or_array.dt.to_pydatetime()
    elif isinstance(series_or_array, (pd.DatetimeIndex, pd.Timestamp)):
        time = series_or_array.to_pydatetime()
    else:
        # convert to us, as ns does return longs instead of datetimes,
        # datetimes cannot handle ns
        time = (
            np.asanyarray(series_or_array)
            .astype('datetime64[us]')
            .astype(datetime)
        )

    return Time(time, scale='utc')


def equatorial_to_camera(ra, dec, zd_pointing, az_pointing, obstime, rotated=True):
    '''
    Convert sky coordinates from the equatorial frame to FACT camera
    coordinates.

    Parameters
    ----------
    ra: number or array-like
        Right ascension in hourangle
    dec: number or array-like
        Declination in degrees
    zd_pointing: number or array-like
        Zenith distance of the telescope pointing direction in degree
    az_pointing: number or array-like
        Azimuth of the telescope pointing direction in degree
    obstime: datetime or np.datetime64
        Time of the observations
    rotated: bool
        True means x points right and y points up when looking on the camera
        from the dish, which is the efinition of FACT-Tools >= 1.0 and Mars.
        False means x points up and y points left,
        which is definition in the original FACTPixelMap file.

    Returns
    -------
    x: number or array-like
        x-coordinate in the camera plane in mm.
        Following the axis of the the FACTPixelMap file (and FACT-Tools).
    y: number or array-like
        y-coordinate in the camera plane in mm.
        Following the axis of the the FACTPixelMap file (and FACT-Tools).
    '''
    eq_coordinates = arrays_to_equatorial(ra, dec, obstime=obstime)
    pointing_direction = arrays_to_altaz(zd_pointing, az_pointing, obstime)

    camera_frame = CameraFrame(pointing_direction=pointing_direction, rotated=rotated)
    cam_coordinates = eq_coordinates.transform_to(camera_frame)

    return cam_coordinates.x.to(u.mm).value, cam_coordinates.y.to(u.mm).value


def camera_to_equatorial(x, y, zd_pointing, az_pointing, obstime, rotated=True):
    '''
    Convert FACT camera coordinates to sky coordinates in the equatorial (icrs)
    frame.

    Parameters
    ----------
    x: number or array-like
        x-coordinate in the camera plane in mm.
        Following the axis of the the FACTPixelMap file (and FACT-Tools).
    y: number or array-like
        y-coordinate in the camera plane in mm.
        Following the axis of the the FACTPixelMap file (and FACT-Tools).
    zd_pointing: number or array-like
        Zenith distance of the telescope pointing direction in degree
    az_pointing: number or array-like
        Azimuth of the telescope pointing direction in degree
    obstime: datetime or np.datetime64
        Time of the observations
    rotated: bool
        True means x points right and y points up when looking on the camera
        from the dish, which is the efinition of FACT-Tools >= 1.0 and Mars.
        False means x points up and y points left,
        which is definition in the original FACTPixelMap file.

    Returns
    -------
    ra: number or array-like
        Right ascension in hourangle
    dec: number or array-like
        Declination in degrees
    '''
    pointing_direction = arrays_to_altaz(zd_pointing, az_pointing, obstime)
    cam_coordinates = arrays_to_camera(
        x, y, pointing_direction, obstime=obstime, rotated=True
    )
    eq_coordinates = cam_coordinates.transform_to(ICRS)

    return eq_coordinates.ra.hourangle, eq_coordinates.dec.deg


def horizontal_to_camera(zd, az, zd_pointing, az_pointing, rotated=True):
    '''
    Convert sky coordinates from the equatorial frame to FACT camera
    coordinates.

    Parameters
    ----------
    zd: number or array-like
        Zenith distance in degrees
    az: number or array-like
        azimuth in degrees
    zd_pointing: number or array-like
        Zenith distance of the telescope pointing direction in degree
    az_pointing: number or array-like
        Azimuth of the telescope pointing direction in degree
    rotated: bool
        True means x points right and y points up when looking on the camera
        from the dish, which is the efinition of FACT-Tools >= 1.0 and Mars.
        False means x points up and y points left,
        which is definition in the original FACTPixelMap file.

    Returns
    -------
    x: number or array-like
        x-coordinate in the camera plane in mm.
        Following the axis of the the FACTPixelMap file (and FACT-Tools).
    y: number or array-like
        y-coordinate in the camera plane in mm.
        Following the axis of the the FACTPixelMap file (and FACT-Tools).
    '''
    altaz = arrays_to_altaz(zd, az)
    pointing_direction = arrays_to_altaz(zd_pointing, az_pointing)

    camera_frame = CameraFrame(
        pointing_direction=pointing_direction, rotated=rotated
    )
    cam_coordinates = altaz.transform_to(camera_frame)

    return cam_coordinates.x.to(u.mm).value, cam_coordinates.y.to(u.mm).value


def camera_to_horizontal(x, y, zd_pointing, az_pointing, rotated=True):
    '''
    Convert FACT camera coordinates to sky coordinates in the equatorial (icrs)
    frame.

    Parameters
    ----------
    x: number or array-like
        x-coordinate in the camera plane in mm.
        Following the axis of the the FACTPixelMap file (and FACT-Tools).
    y: number or array-like
        y-coordinate in the camera plane in mm.
        Following the axis of the the FACTPixelMap file (and FACT-Tools).
    zd_pointing: number or array-like
        Zenith distance of the telescope pointing direction in degree
    az_pointing: number or array-like
        Azimuth of the telescope pointing direction in degree
    rotated: bool
        True means x points right and y points up when looking on the camera
        from the dish, which is the efinition of FACT-Tools >= 1.0 and Mars.
        False means x points up and y points left,
        which is definition in the original FACTPixelMap file.

    Returns
    -------
    zd: number or array-like
        Zenith distance in degrees
    az: number or array-like
        Declination in degrees
    '''
    pointing_direction = arrays_to_altaz(zd_pointing, az_pointing)
    cam_coordinates = arrays_to_camera(
        x, y, pointing_direction, rotated=rotated
    )
    altaz = cam_coordinates.transform_to(AltAz(location=LOCATION))

    return altaz.zen.deg, altaz.az.deg



from setuptools import setup, find_packages

with open('fact/VERSION', 'r') as f:
    __version__ = f.read().strip()

with open('README.rst') as f:
    long_description = f.read()


setup(
    name='pyfact',
    version=__version__,
    description='A module containing useful methods for working with fact',
    long_description=long_description,
    url='http://github.com/fact-project/pyfact',
    author='Maximilian Noethe, Dominik Neise',
    author_email='maximilian.noethe@tu-dortmund.de',
    license='MIT',
    packages=find_packages(),
    package_data={
        '': [
            'VERSION',
            'resources/*',
            'credentials/credentials.encrypted',
        ]
    },
    entry_points={
        'console_scripts': [
            'fact_calculate_theta = fact.analysis.scripts.theta:main',
            'fact_calculate_radec = fact.analysis.scripts.radec:main',
        ]
    },
    tests_require=['pytest>=3.0.0'],
    setup_requires=['pytest-runner'],
    install_requires=[
        'astropy',
        'h5py',
        'matplotlib>=1.4',
        'numpy',
        'pandas',
        'peewee==2.*',
        'pymongo>=2.7',
        'pymysql',
        'python-dateutil',
        'scipy',
        'setuptools',
        'simple-crypt',
        'sqlalchemy',
        'tables',  # pytables in anaconda
        'wrapt',
    ],
    zip_safe=False,
)



import numpy as np
import astropy.units as u

POINT_SOURCE_FLUX_UNIT = (1 / u.GeV / u.s / u.m**2).unit
FLUX_UNIT = POINT_SOURCE_FLUX_UNIT / u.sr


@u.quantity_input
def random_power(
    spectral_index,
    e_min: u.TeV,
    e_max: u.TeV,
    size,
    e_ref: u.TeV=1 * u.TeV,
) -> u.TeV:
    r'''
    Draw random numbers from a power law distribution

    .. math::
        f(E) =
        \frac{1 - \gamma}
        ({E_{\max} / E_{\mathrm{ref}})^{\gamma - 1} - (E_{\min} / E_{\mathrm{ref}})^{\gamma - 1}}
        (E /  E_{\mathrm{ref}})^{\gamma}

    Parameters
    ----------
    spectral_index: float
        The differential spectral index of the power law
    e_min: Quantity[energy]
        lower energy border
    e_max: Quantity[energy]
        upper energy border, can be np.inf
    size: int or tuple[int]
        Number of events to draw or a shape like (100, 2)
    '''
    if spectral_index > -1.0:
        raise ValueError('spectral_index must be < -1.0')

    u = np.random.uniform(0, 1, size)

    exponent = spectral_index + 1

    if e_max == np.inf:
        diff = -(e_min / e_ref)**exponent
    else:
        diff = (e_max / e_ref)**exponent - (e_min / e_ref)**exponent

    return e_ref * (diff * u + (e_min / e_ref)**exponent) ** (1 / exponent)


def power_law(
    energy: u.TeV,
    flux_normalization: (FLUX_UNIT, POINT_SOURCE_FLUX_UNIT),
    spectral_index,
    e_ref: u.TeV=1 * u.TeV,
):
    r'''
    Simple power law

    .. math::
        \phi = \phi_0 \cdot (E / E_{}\mathrm{ref}})^{\gamma}

    Parameters
    ----------
    energy: Quantity[energy]
        energy points to evaluate
    flux_normalization: Quantity[m**-2 s**-2 TeV**-1]
        Flux normalization
    spectral_index: float
        Spectral index
    e_ref: Quantity[energy]
        The reference energy
    '''
    return flux_normalization * (energy / e_ref)**(spectral_index)


@u.quantity_input
def curved_power_law(
    energy: u.TeV,
    flux_normalization: (FLUX_UNIT, POINT_SOURCE_FLUX_UNIT),
    a,
    b,
    e_ref: u.TeV=1 * u.TeV,
):
    r'''
    Curved power law

    .. math::
        \phi = \phi_0 \cdot E ^ {a + b \cdot \log_{10}(E)}

    Parameters
    ----------
    energy: number or array-like
        energy points to evaluate
    flux_normalization: float
        Flux normalization
    a: float
        Parameter `a`  of the curved power law
    b: float
        Parameter `b`  of the curved power law
    e_ref: Quantity[energy]
        The reference energy
    '''
    exp = a + b * np.log10((energy / e_ref))
    return flux_normalization * (energy / e_ref) ** exp


def li_ma_significance(n_on, n_off, alpha=0.2):
    '''
    Calculate the Li&Ma significance for given
    observations data

    Parameters
    ----------
    n_on: integer or array like
        Number of events for the on observations
    n_off: integer of array like
        Number of events for the off observations
    alpha: float
        Scaling factor for the off observations, for wobble observations
        this is 1 / number of off regions
    '''

    scalar = np.isscalar(n_on)

    n_on = np.array(n_on, copy=False, ndmin=1)
    n_off = np.array(n_off, copy=False, ndmin=1)

    with np.errstate(divide='ignore', invalid='ignore'):
        p_on = n_on / (n_on + n_off)
        p_off = n_off / (n_on + n_off)

        t1 = n_on * np.log(((1 + alpha) / alpha) * p_on)
        t2 = n_off * np.log((1 + alpha) * p_off)

        ts = (t1 + t2)
        significance = np.sqrt(ts * 2)

    significance[np.isnan(significance)] = 0
    significance[n_on < alpha * n_off] = 0

    if scalar:
        return significance[0]

    return significance


@u.quantity_input
def calc_weight_simple_to_curved(
    energy: u.TeV,
    spectral_index,
    a,
    b,
    e_ref: u.TeV = 1 * u.TeV,
):
    '''
    Reweight simulated events from a simulated power law with
    spectral index `spectral_index` to a curved power law with parameters `a` and `b`

    Parameters
    ----------
    energy: float or array-like
        Energy of the events
    spectral_index: float
        Spectral index of the simulated power law
    a: float
        Parameter `a` of the target curved power law
    b: float
        Parameter `b` of the target curved power law
    '''
    return (energy / e_ref) ** (spectral_index + a + b * np.log10(energy / e_ref))


@u.quantity_input
def calc_weight_change_index(
    energy: u.TeV,
    simulated_index,
    target_index,
    e_ref: u.TeV=1 * u.TeV,
):
    '''
    Reweight simulated events from one power law index to another

    Parameters
    ----------
    energy: float or array-like
        Energy of the events
    simulated_index: float
        Spectral index of the simulated power law
    target_index: float
        Spectral index of the target power law
    '''
    return (energy / e_ref) ** (target_index - simulated_index)


@u.quantity_input
def calc_gamma_obstime(
    n_events,
    spectral_index,
    flux_normalization: (FLUX_UNIT, POINT_SOURCE_FLUX_UNIT),
    max_impact: u.m,
    e_min: u.TeV,
    e_max: u.TeV,
    e_ref: u.TeV = 1 * u.TeV,
) -> u.s:
    r'''
    Calculate the equivalent observation time for a number of simulation enents

    The number of events produced by sampling from a power law with
    spectral index γ is given by

    .. math::
        N =
        A t \Phi_0 \int_{E_{\min}}^{E_{\max}}
        \left(\frac{E}{E_{\mathrm{ref}}}\right)^{\gamma}
        \,\mathrm{d}E

    Solving this for t yields

    .. math::
        t = \frac{N \cdot (\gamma - 1)}{
            E_{\mathrm{ref}} A \Phi_0
            \left(
                \left(\frac{E_{\max}}{E_{\mathrm{ref}}}\right)^{\gamma - 1}
                - \left(\frac{E_{\max}}{E_{\mathrm{ref}}}\right)^{\gamma - 1}
            \right)
        }

    Parameters
    ----------
    n_events: int
        Number of simulated events
    spectral_index: float
        Spectral index of the simulated power law, including the sign,
        so typically -2.7 or -2
    flux_normalization: float
        Flux normalization of the simulated power law
    max_impact: Quantity[length]
        Maximal simulated impact
    e_min: Quantity[energy]
        Mimimal simulated energy
    e_max: Quantity[energy]
        Maximal simulated energy
    e_ref: Quantity[energy]
        The e_ref energy
    '''
    if spectral_index >= -1:
        raise ValueError('spectral_index must be < -1')

    numerator = n_events * (spectral_index + 1)

    A = max_impact**2 * np.pi
    t1 = A * flux_normalization * e_ref
    t2 = (e_max / e_ref)**(spectral_index + 1)
    t3 = (e_min / e_ref)**(spectral_index + 1)

    denominator = t1 * (t2 - t3)

    return numerator / denominator


@u.quantity_input
def power_law_integral(
    flux_normalization: (FLUX_UNIT, POINT_SOURCE_FLUX_UNIT),
    spectral_index,
    e_min: u.TeV,
    e_max: u.TeV,
    e_ref: u.TeV = 1 * u.TeV,
):
    '''
    Return the integral of a power_law with normalisation
    `flux_normalization` and index `spectral_index`
    between `e_min` and `e_max`
    '''
    int_index = spectral_index + 1
    e_term = (e_max / e_ref)**(int_index) - (e_min / e_ref)**(int_index)

    res = flux_normalization * e_ref / int_index * e_term

    if flux_normalization.unit.is_equivalent(FLUX_UNIT):
        return res.to(1 / u.m**2 / u.s / u.sr)
    return res.to(1 / u.m**2 / u.s)


@u.quantity_input
def calc_proton_obstime(
    n_events,
    spectral_index,
    max_impact: u.m,
    viewcone: u.deg,
    e_min: u.TeV,
    e_max: u.TeV,
    flux_normalization: FLUX_UNIT = 1.8e4 * FLUX_UNIT,
    e_ref: u.GeV=1 * u.GeV,
) -> u.s:
    '''
    Calculate the equivalent observation time for a proton montecarlo set

    Parameters
    ----------
    n_events: int
        Number of simulated events
    spectral_index: float
        Spectral index of the simulated power law
    max_impact: float
        Maximal simulated impact in m
    viewcone: float
        Viewcone in degrees
    e_min: Quantity[energy]
        Mimimal simulated energy
    e_max: Quantity[energy]
        Maximal simulated energy
    flux_normalization: float
        Flux normalisation of the cosmic rays in nucleons / (m² s sr GeV)
        Default value is (29.2) of
        http://pdg.lbl.gov/2016/reviews/rpp2016-rev-cosmic-rays.pdf
    '''
    area = np.pi * max_impact**2
    solid_angle = 2 * np.pi * (1 - np.cos(viewcone)) * u.sr

    expected_integral_flux = power_law_integral(
        flux_normalization, spectral_index, e_min, e_max, e_ref
    )

    return n_events / area / solid_angle / expected_integral_flux



from functools import lru_cache
from configparser import SafeConfigParser
from getpass import getpass
from simplecrypt import decrypt
from io import StringIO
from pkg_resources import resource_stream
from sqlalchemy import create_engine
import socket
import os


__all__ = ['get_credentials', 'create_factdb_engine']


@lru_cache(1)
def get_credentials():
    '''
    Get a SafeConfigParser instance with FACT credentials
    On the first call, you will be prompted for the FACT password

    The folling credentials are stored:

    - telegram
        - token

    - database
        - user
        - password
        - host
        - database

    - twilio
        - sid
        - auth_token
        - number

    use get_credentials().get(group, element) to retrieve elements
    '''
    if 'FACT_PASSWORD' in os.environ:
        passwd = os.environ['FACT_PASSWORD']
    else:
        passwd = getpass('Please enter the current, universal FACT password: ')

    with resource_stream('fact', 'credentials/credentials.encrypted') as f:
        decrypted = decrypt(passwd, f.read()).decode('utf-8')

    config = SafeConfigParser()
    config.readfp(StringIO(decrypted))

    return config


def create_factdb_engine(database=None):
    '''
    returns a sqlalchemy.Engine pointing to the factdata database

    The different hostname on isdc machines is handled correctly.
    '''
    spec = 'mysql+pymysql://{user}:{password}@{host}/{database}'

    creds = get_credentials()
    config = dict(creds['database'])

    if socket.gethostname().startswith('isdc'):
        config['host'] = 'lp-fact'

    if database is not None:
        config['database'] = database

    return create_engine(
        spec.format(**config),
        connect_args={'ssl': {'ssl-mode': 'preferred'}},
    )



from .database import factdata_db, connect_database
from .utils import (
    read_into_dataframe,
    get_ontime_by_source,
    get_ontime_by_source_and_runtype,
    get_correct_ontime,
    ontime,
    run_duration,
)
from .models import *



import numpy as np

import astropy.units as u
from astropy.coordinates import AltAz, SkyCoord
from astropy.coordinates.angle_utilities import angular_separation

from ..coordinates import CameraFrame
from ..coordinates.utils import (
    arrays_to_camera,
    arrays_to_altaz,
    arrays_to_equatorial,
)
from ..instrument.constants import LOCATION


def calc_off_position(source_x, source_y, off_index, n_off=5):
    '''
    For a given source position in camera coordinates,
    return an array with n_off off_positions

    Parameters
    ----------
    source_x: float or astropy.units.Quantity
        x position of the wanted source position
    source_y: float or astropy.units.Quantity
        y position of the wanted source position
    n_off: int
        number of off positions to calculate

    Returns
    -------
    x_off: float or astropy.units.Quantity
        x coordinate of the off position
    y_off: float or astropy.units.Quantity
        y coordinate of the off position
    '''

    if off_index < 1 or off_index > n_off:
        raise ValueError('off_index must be >= 1 and <= n_off')

    r = np.sqrt(source_x**2 + source_y**2)
    phi = np.arctan2(source_y, source_x)
    delta_phi = 2 * np.pi / (n_off + 1)
    if hasattr(phi, 'unit'):
        delta_phi *= u.rad

    x_off = r * np.cos(phi + off_index * delta_phi)
    y_off = r * np.sin(phi + off_index * delta_phi)

    return x_off, y_off


def calc_theta_equatorial(
        source_ra_prediction,
        source_dec_prediction,
        source_ra,
        source_dec):
    '''
    Calculate the angular distance between reconstructed source
    position and assumed source position, both given in the
    equatorial coordinate system.

    Parameters
    ----------
    source_ra_prediction: number or array-like
        prediction of the right ascension of the source position
        in hourangle
    source_dec_prediction: number or array-like
        prediction of the declination of the source position
        in degree
    source_ra: number or array-like
        Right ascension of the source position in hourangle
    source_dec: number or array-like
        Declination of the source position in degree

    Returns
    -------
    theta_deg: array
        theta in degrees
    '''
    source_prediction = arrays_to_equatorial(
        source_ra_prediction, source_dec_prediction,
    )
    source_pos = arrays_to_equatorial(source_ra, source_dec)

    return source_pos.separation(source_prediction).deg


def calc_theta_camera(
        source_x_prediction,
        source_y_prediction,
        source_zd,
        source_az,
        zd_pointing,
        az_pointing):
    '''
    Calculate the angular distance between reconstructed source
    position and assumed source position, where the
    prediction is given as camera coordinates and the source position
    in the horizontal coordinate system.

    Parameters
    ----------
    source_x_prediction: number or array-like
        prediction of the x postions of the source in mm
    source_y_prediction: number or array-like
        prediction of the y postions of the source in mm
    source_zd: number or array-like
        Zenith of the source position in degree
    source_az: number or array-like
        Azimuth of the source position in degree
    zd_pointing: number or array-like
        zenith angle of the pointing direction in degrees
    az_pointing: number or array-like
        azimuth angle of the pointing direction in degrees

    Returns
    -------
    theta_deg: array
        theta in degrees
    '''
    pointing = arrays_to_altaz(zd_pointing, az_pointing)
    altaz = AltAz(location=LOCATION)

    source_prediction = arrays_to_camera(
        source_x_prediction, source_y_prediction,
        pointing_direction=pointing,
    )

    source_pos = arrays_to_altaz(source_zd, source_az)
    source_prediction_alt_az = source_prediction.transform_to(altaz)
    return angular_separation(
        source_prediction_alt_az.az, source_prediction_alt_az.alt,
        source_pos.az, source_pos.alt,
    ).to(u.deg).value


def calc_theta_offs_camera(
        source_x_prediction,
        source_y_prediction,
        source_zd,
        source_az,
        zd_pointing,
        az_pointing,
        n_off=5):
    '''
    Calculate the angular distance between reconstructed source
    position and `n_off` off positions.

    Parameters
    ----------
    source_x_prediction: number or array-like
        prediction of the x coordinate of the source position in mm
    source_y_prediction: number or array-like
        prediction of the y coordinate of the source position in mm
    source_zd: number or array-like
        Zenith of the source position in degree
    source_az: number or array-like
        Azimuth of the source position in degree
    zd_pointing: number or array-like
        zenith angle of the pointing direction in degrees
    az_pointing: number or array-like
        azimuth angle of the pointing direction in degrees
    n_off: int
        How many off positions to calculate

    Returns
    -------
    theta_deg: n_off-tuple
        theta in degrees for each off position
    '''
    pointing = arrays_to_altaz(zd_pointing, az_pointing)
    source = arrays_to_altaz(source_zd, source_az)
    source_prediction = arrays_to_camera(
        source_x_prediction, source_y_prediction,
        pointing_direction=pointing,
    )

    altaz = AltAz(location=LOCATION)
    camera_frame = CameraFrame(
        location=LOCATION,
        pointing_direction=pointing
    )

    source_prediction_alt_az = source_prediction.transform_to(altaz)
    source_camera = source.transform_to(camera_frame)

    theta_offs = []

    for i in range(1, n_off + 1):
        off_x, off_y = calc_off_position(
            source_camera.x, source_camera.y, off_index=i, n_off=n_off
        )

        off_pos = SkyCoord(off_x, off_y, frame=camera_frame)
        off_pos_altaz = off_pos.transform_to(altaz)
        theta_offs.append(angular_separation(
            source_prediction_alt_az.az, source_prediction_alt_az.alt,
            off_pos_altaz.az, off_pos_altaz.alt,
        ).to(u.deg).value)
    return tuple(theta_offs)
